// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/16AIO/16AIO_Linux_4.x.x.x_DN/driver/read.c $
// $Rev: 51278 $
// $Date: 2022-07-07 16:51:21 -0500 (Thu, 07 Jul 2022) $

// 16AIO: Device Driver: source file

#include "main.h"



/******************************************************************************
*
*	Function:	_get_fill_level
*
*	Purpose:
*
*		Determin the fill level of a FIFO. We do this via a binary search since
*		the threshold flag dynamically adjusts to the threshold value and the
*		FIFO fill level.
*
*		NOTE: The threshold flag may not respond appropriately if the FIFO is
*				full (32768 values) when the threshold level set to 0x7FFF
*				(32767). This routine therefore returns a value of 32767 even
*				when the FIFO is full.
*
*	Arguments:
*
*		dev		The structure for the device to access.
*
*	Returned:
*
*		>= 0	This is the nunmber of values in the FIFO and the transfer
*				buffer.
*
******************************************************************************/

static long _get_fill_level(dev_data_t* dev)
{
	u32		icr;
	u32		max	= 0x7FFF;
	u32		mid;
	u32		min	= 0x0000;
	u32		new;
	u32		obcr;
	long	qty	= 0;
	VADDR_T	vaddr	= dev->vaddr.gsc_ibcr_32;

	// Fix the interrupt option while we use the threshold value.
	icr		= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_icr_32);

	if (((icr & 0x700) == 0x100) || ((icr & 0x700) == 0x200))
		os_reg_mem_mx_u32(dev, dev->vaddr.gsc_icr_32, 0x700, 0x000);

	// Record the original threshold value;
	obcr	= os_reg_mem_rx_u32(dev, vaddr);

	// Do a binary search to get the buffer's fill level.
	for (;;)
	{
		mid	= (min + max) / 2;
		os_reg_mem_tx_u32(dev, vaddr, mid);
		new	= os_reg_mem_rx_u32(dev, vaddr);

		if (new & 0x10000)
		{
			// The level exceeds the value of mid.
			min	= mid + 1;

			if (min > max)
			{
				qty	= min;
				break;
			}
		}
		else
		{
			// The level meets or falls below the value of mid.

			if (min == max)
			{
				qty	= min;
				break;
			}

			max	= mid;
		}
	}

	// Restore the original threshold value;
	os_reg_mem_tx_u32(dev, vaddr, obcr);

	// Restore the interrupt option.

	if (((icr & 0x700) == 0x100) || ((icr & 0x700) == 0x200))
		os_reg_mem_mx_u32(dev, dev->vaddr.gsc_icr_32, 0x700, icr);

	return(qty);
}



//*****************************************************************************
static void _dev_io_sw_init(dev_data_t* dev, dev_io_t* io)
{
	io->io_mode			= AIO_IO_MODE_DEFAULT;
	io->pio_threshold	= 32;
	io->timeout_s		= AIO_IO_TIMEOUT_DEFAULT;
}



//*****************************************************************************
static void _dev_io_close(dev_data_t* dev, dev_io_t* io)
{
	io->dma_channel	= NULL;
}



//*****************************************************************************
static void _dev_io_open(dev_data_t* dev, dev_io_t* io)
{
	_dev_io_sw_init(dev, io);
}



//*****************************************************************************
static int _dev_io_startup(dev_data_t* dev, dev_io_t* io)
{
	// We've nothing to check here.
	return(0);
}



//*****************************************************************************
static long _dev_pio_available(dev_data_t* dev, dev_io_t* io, size_t bytes)
{
	bytes	= _get_fill_level(dev);
	bytes	*= 4;
	return(bytes);
}



//*****************************************************************************
static long _dev_pio_xfer(
	dev_data_t*		dev,
	dev_io_t*		io,
	const os_mem_t*	mem,
	size_t			bytes,
	os_time_tick_t	st_end)
{
	long	qty;

	qty	= gsc_read_pio_work_32_bit(dev, io, mem, bytes, st_end);
	return(qty);
}



//*****************************************************************************
static long _dev_bmdma_available(dev_data_t* dev, dev_io_t* io, size_t bytes)
{
	u32	ibcr;

	ibcr	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_ibcr_32);

	if (ibcr & 0x10000)
	{
		// The fill level is above the threshold.
		bytes	= (ibcr & 0x7FFF) * 4;
	}
	else
	{
		// The input buffer has too little data.
		bytes	= 0;
	}

	return(bytes);
}



//*****************************************************************************
static long _dev_bmdma_xfer(
	dev_data_t*		dev,
	dev_io_t*		io,
	const os_mem_t*	mem,
	size_t			bytes,
	os_time_tick_t	st_end)
{
	long			qty;
	long			samples	= bytes / 4;
	gsc_dma_setup_t	setup;

	if (samples < io->pio_threshold)
	{
		qty	= gsc_read_pio_work_32_bit(dev, io, mem, bytes, st_end);
	}
	else
	{
		memset(&setup, 0, sizeof(gsc_dma_setup_t));
		setup.alt		= dev;
		setup.dev		= dev;
		setup.io		= io;
		setup.mem		= mem;
		setup.st_end	= st_end;
		setup.bytes		= bytes;
		setup.ability	= GSC_DMA_CAP_BMDMA_READ;

		setup.mode		= GSC_DMA_MODE_BLOCK_DMA
						| GSC_DMA_MODE_SIZE_32_BITS
						| GSC_DMA_MODE_INPUT_ENABLE
						| GSC_DMA_MODE_BURSTING_LOCAL
						| GSC_DMA_MODE_INTERRUPT_WHEN_DONE
						| GSC_DMA_MODE_LOCAL_ADRESS_CONSTANT
						| GSC_DMA_MODE_PCI_INTERRUPT_ENABLE;

		setup.dpr		= GSC_DMA_DPR_BOARD_TO_HOST
						| GSC_DMA_DPR_END_OF_CHAIN
						| GSC_DMA_DPR_TERMINAL_COUNT_IRQ;

		qty	= gsc_dma_perform(&setup);
	}

	return(qty);
}



//*****************************************************************************
static long _dev_dmdma_available(dev_data_t* dev, dev_io_t* io, size_t bytes)
{
	// DMDMA is unsupported.
	return(-EINVAL);
}



//*****************************************************************************
static long _dev_dmdma_xfer(
	dev_data_t*		dev,
	dev_io_t*		io,
	const os_mem_t*	mem,
	size_t			bytes,
	os_time_tick_t	st_end)
{
	// DMDMA is unsupported.
	return(-EINVAL);
}



/******************************************************************************
*
*	Function:	dev_read_create
*
*	Purpose:
*
*		Perform a one-tine setup of the Read Analog Input streaming I/O structure.
*
*	Arguments:
*
*		dev		The data for the device of interest.
*
*		io		The I/O structure for this stream.
*
*	Returned:
*
*		None.
*
******************************************************************************/

int dev_read_create(dev_data_t* dev, dev_io_t* io)
{
	int	ret;

	io->bytes_per_sample	= 4;
	io->io_reg_offset		= GSC_REG_OFFSET(AIO_GSC_IDBR);
	io->io_reg_vaddr		= dev->vaddr.gsc_idbr_32;

	io->dev_io_sw_init		= _dev_io_sw_init;
	io->dev_io_close		= _dev_io_close;
	io->dev_io_open			= _dev_io_open;
	io->dev_io_startup		= _dev_io_startup;
	io->dev_pio_available	= _dev_pio_available;
	io->dev_bmdma_available	= _dev_bmdma_available;
	io->dev_dmdma_available	= _dev_dmdma_available;
	io->dev_pio_xfer		= _dev_pio_xfer;
	io->dev_bmdma_xfer		= _dev_bmdma_xfer;
	io->dev_dmdma_xfer		= _dev_dmdma_xfer;

	io->wait.abort			= AIO_WAIT_IO_RX_ABORT;
	io->wait.done			= AIO_WAIT_IO_RX_DONE;
	io->wait.error			= AIO_WAIT_IO_RX_ERROR;
	io->wait.timeout		= AIO_WAIT_IO_RX_TIMEOUT;

	ret	= gsc_io_create(dev, io, 32L * 1024 * 4);

	return(ret);
}


